{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "#--------------------------------------------------------------------\n",
    "# TODO: make sure to use my latest `advent` lib instead of this inline BS\n",
    "class GrowableGrid():\n",
    "    def __init__(self, start_node):\n",
    "        self.rows = 1\n",
    "        self.cols = 1\n",
    "        self.start_row = 0\n",
    "        self.start_col = 0\n",
    "        self.row = 0\n",
    "        self.col = 0\n",
    "        self.grid = [[start_node]]\n",
    "        self.stack = []\n",
    "\n",
    "    def create_row(self):\n",
    "        return [None] * self.cols\n",
    "\n",
    "    def create_col(self):\n",
    "        return [None] * self.cols\n",
    "\n",
    "    def assert_not_set(self, row, col, node):\n",
    "        existing_node = self.grid[row][col]\n",
    "        if existing_node != node and existing_node is not None:\n",
    "            raise Exception(\n",
    "                f\"Trying to insert {node} into (row={row}, col={col})\" +\n",
    "                f\" but it already contains {self.grid[row][col]}\")\n",
    "\n",
    "    def up(self, node):\n",
    "        if self.row == 0:\n",
    "            new_row = self.create_row()\n",
    "            self.grid.insert(0, new_row)\n",
    "            self.rows += 1\n",
    "            self.recalc_positions(1, 0)\n",
    "\n",
    "        self.assert_not_set(self.row-1, self.col, node)\n",
    "        self.grid[self.row-1][self.col] = node\n",
    "        self.row -= 1\n",
    "\n",
    "    def down(self, node):\n",
    "        if self.row == self.rows - 1:\n",
    "            new_row = self.create_row()\n",
    "            self.grid.append(new_row)\n",
    "            self.rows += 1\n",
    "\n",
    "        self.assert_not_set(self.row+1, self.col, node)\n",
    "        self.grid[self.row+1][self.col] = node\n",
    "        self.row += 1\n",
    "\n",
    "    def left(self, node):\n",
    "        if self.col == 0:\n",
    "            for row in self.grid:\n",
    "                row.insert(0, None)\n",
    "            self.recalc_positions(0, 1)\n",
    "            self.cols += 1\n",
    "            \n",
    "        self.assert_not_set(self.row, self.col-1, node)\n",
    "        self.grid[self.row][self.col-1] = node\n",
    "        self.col -= 1\n",
    "\n",
    "    def right(self, node):\n",
    "        if self.col == self.cols - 1:\n",
    "            for row in self.grid:\n",
    "                row.append(None)\n",
    "            self.cols += 1\n",
    "\n",
    "        self.assert_not_set(self.row, self.col+1, node)\n",
    "        self.grid[self.row][self.col+1] = node\n",
    "        self.col += 1\n",
    "        \n",
    "    def recalc_positions(self, d_row, d_col):\n",
    "        self.row += d_row\n",
    "        self.col += d_col\n",
    "        self.start_row += d_row\n",
    "        self.start_col += d_col\n",
    "        \n",
    "        for i in range(len(self.stack)):\n",
    "            row, col = self.stack[i]\n",
    "            self.stack[i] = (row + d_row, col + d_col)\n",
    "\n",
    "    def push_position(self):\n",
    "        self.stack.append((self.row, self.col))\n",
    "    \n",
    "    def pop_position(self):\n",
    "        row, col = self.stack.pop()\n",
    "        self.row = row\n",
    "        self.col = col\n",
    "\n",
    "    def peek_position(self):\n",
    "        row, col = self.stack[-1]\n",
    "        self.row = row\n",
    "        self.col = col\n",
    "\n",
    "    def __str__(self):\n",
    "        rs = []\n",
    "        for row in self.grid:\n",
    "            row_str = []\n",
    "            for col in row:\n",
    "                if col:\n",
    "                    row_str.append(col)\n",
    "                else:\n",
    "                    row_str.append(\" \")\n",
    "            rs.append(\"\".join(row_str))\n",
    "\n",
    "        return \"\\n\".join(rs)\n",
    "\n",
    "def neighbours4(grid, r, c):\n",
    "    \"\"\"\n",
    "        Returns the adjacent tiles NOT including diagonals\n",
    "    \"\"\"\n",
    "    max_r = len(grid)\n",
    "    max_c = len(grid[0])\n",
    "    candidates = (\n",
    "        (r-1, c), (r+1, c),\n",
    "        (r, c-1), (r, c+1)\n",
    "    )\n",
    "    for maybe_r, maybe_c in candidates:\n",
    "        if maybe_r == r and maybe_c == c:\n",
    "            continue\n",
    "        if maybe_r >= max_r or maybe_r < 0:\n",
    "            continue\n",
    "        if maybe_c >= max_c or maybe_c < 0:\n",
    "            continue\n",
    "\n",
    "        yield (maybe_r, maybe_c)\n",
    "\n",
    "def expand_grid(i, regex_str, grid):\n",
    "    grid.push_position()\n",
    "    \n",
    "    while i < len(regex_str):\n",
    "        c = regex_str[i]\n",
    "        if c == \"(\":\n",
    "            r, c = (grid.row, grid.col)\n",
    "            i = expand_grid(i+1, regex_str, grid)\n",
    "            grid.row, grid.col = r, c\n",
    "        elif c == \")\":\n",
    "            grid.pop_position()\n",
    "            return i\n",
    "        elif c == \"|\":\n",
    "            grid.peek_position()\n",
    "        else:\n",
    "            if c == \"N\":\n",
    "                grid.up(\"-\")\n",
    "                grid.up(\".\")\n",
    "            elif c == \"S\":\n",
    "                grid.down(\"-\")\n",
    "                grid.down(\".\")\n",
    "            elif c == \"E\":\n",
    "                grid.right(\"|\")\n",
    "                grid.right(\".\")\n",
    "            elif c == \"W\":\n",
    "                grid.left(\"|\")\n",
    "                grid.left(\".\")\n",
    "        i += 1\n",
    "    return i        "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "part 1: 3755\n",
      "part 2: 8627\n"
     ]
    }
   ],
   "source": [
    "def solve(input_str):\n",
    "    g = GrowableGrid(\"X\")\n",
    "    expand_grid(0, input_str, g)\n",
    "    \n",
    "    # make move grid\n",
    "    move_grid = []\n",
    "    for row in g.grid:\n",
    "        move_grid.append([ None for _ in row ])\n",
    "        \n",
    "    move_grid[g.start_row][g.start_col] = 0\n",
    "    queue = [(g.start_row, g.start_col)]\n",
    "    doors = 1\n",
    "    \n",
    "    while any(queue):\n",
    "        next_queue = []\n",
    "        \n",
    "        for r,c in queue:\n",
    "            neighbours = neighbours4(g.grid, r, c)\n",
    "            for nr, nc in neighbours:\n",
    "                if g.grid[nr][nc] in (\"-\", \"|\", \".\") and move_grid[nr][nc] == None:\n",
    "                    dr = nr - r\n",
    "                    dc = nc - c\n",
    "                    next_queue.append((r + dr, c + dc))\n",
    "                \n",
    "                if move_grid[nr][nc] == None or move_grid[nr][nc] > doors:\n",
    "                    if g.grid[nr][nc] == \".\":\n",
    "                        move_grid[nr][nc] = doors\n",
    "                    else:\n",
    "                        move_grid[nr][nc] = -1\n",
    "        doors += 1\n",
    "        queue = next_queue\n",
    "\n",
    "    print(\"part 1:\", (doors // 2) - 1)\n",
    "    \n",
    "    at_least_1k = 0\n",
    "    for r, row in enumerate(g.grid):\n",
    "        for c, col in enumerate(row):\n",
    "            if col == \".\":\n",
    "                distance = move_grid[r][c]\n",
    "                if distance//2 >= 1000:\n",
    "                    at_least_1k += 1\n",
    "    \n",
    "    print(\"part 2:\", at_least_1k)\n",
    "\n",
    "puzzle_input_str = open(\"puzzle_input/day20.txt\").read()\n",
    "solve(puzzle_input_str)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
